// Copyright ram-0000
// Je donne mon accord  Developpez.com pour une ventuelle intgration dans les pages Sources du site.

#include "Grille.h"
#include <iostream>

CGrille::CGrille(void)
{
	// le constructeur construit une grille vide
	Vide();
}

// operateur de copie
CGrille & CGrille::operator = (const CGrille & Other)
{
	memcpy(m_case, Other.m_case, sizeof(m_case));
	m_far_win = Other.m_far_win;
	m_hash = Other.m_hash;
	m_move_win = Other.m_move_win;

	return *this;
}

bool CGrille::operator == (const CGrille & Other) const
{
	return m_hash == Other.m_hash;
}

bool CGrille::operator < (const CGrille & Other) const
{
	return m_hash < Other.m_hash;
}

// remet la grille a 0
void CGrille::Vide(void)
{
	memset(m_case, 0, sizeof(m_case));
	m_far_win = -1;
	_compute_hash();
}

bool CGrille::Load(std::ifstream & InputStream)
{
	InputStream.read((char *)m_case, sizeof(m_case));
	InputStream.read((char *)&m_far_win, sizeof(m_far_win));
	m_move_win.Load(InputStream);

	_compute_hash();

	return true;
}

bool CGrille::Save(std::ofstream & OutputStream) const
{
	OutputStream.write((char *)m_case, sizeof(m_case));
	OutputStream.write((char *)&m_far_win, sizeof(m_far_win));
	m_move_win.Save(OutputStream);

	return true;
}

std::string CGrille::Dump(void) const
{
	std::string ret;
	for(int boucle = 0; boucle != sizeof(m_case); boucle++)
	{
		switch(m_case[boucle])
		{
		case 0 : ret += "."; break;
		case 1 : ret += "1"; break;
		case 2 : ret += "2"; break;
		case 3 : ret += "3"; break;
		case 4 : ret += "4"; break;
		}
	}
	return ret;
}

void CGrille::Dump(CLogger & Log) const
{
	for(int boucle = 0; boucle != sizeof(m_case); boucle++)
	{
		const char *carac = "?";
		switch(m_case[boucle])
		{
		case 0 : carac = " "; break;
		case 1 : carac = "1"; break;
		case 2 : carac = "2"; break;
		case 3 : carac = "3"; break;
		case 4 : carac = "4"; break;
		}

		Log << carac;

		if( (boucle == 3) || (boucle == 7) || (boucle == 11) || (boucle == 15) || (boucle == 19) )
			Log << "\n";
	}
	Log << "\n";
}

void CGrille::GeneratePetitCarre(MapGrille_t & map_output) const
{
	// une grille temporaire
	CGrille tmp;
	tmp = *this;

	// cherche pour toutes les positions possible si on peut mettre le carre
	for(int pos = 0; pos != sizeof(m_case); pos++)
	{
		if(IsFree(pos) == true)
		{
			// oui, on peut le mettre
			// modifie l'etat des cases
			tmp.SetPetitCarre(pos);

			// enregistre la position
			map_output.insert(MapPair_t(tmp.GetHash(), tmp));

			// remet la grille a sa position d'origine
			tmp = *this;
		}
	}
}

void CGrille::GenerateGrosCarre(MapGrille_t & map_output) const
{
	// une grille temporaire
	CGrille tmp;
	tmp = *this;

// la liste de toutes les positions possibles du carre 2X2
static int list_pos[][4] =
{ {  0,  1,  4,  5 },   {  1,  2,  5,  6 },   {  2,  3,  6,  7 },   {  4,  5,  8,  9 },
  {  5,  6,  9, 10 },   {  6,  7, 10, 11 },   {  8,  9, 12, 13 },   {  9, 10, 13, 14 },
  { 10, 11, 14, 15 },   { 12, 13, 16, 17 },   { 13, 14, 17, 18 },   { 14, 15, 18, 19 },
};

	// cherche pour toutes les positions possible si on peut mettre le carre
	for(int ligne = 0; ligne != sizeof(list_pos) / sizeof(list_pos[0]); ligne++)
	{
		if( (IsFree(list_pos[ligne][0]) == true) &&
			(IsFree(list_pos[ligne][1]) == true) &&
			(IsFree(list_pos[ligne][2]) == true) &&
			(IsFree(list_pos[ligne][3]) == true) )
		{
			// oui, on peut le mettre
			// modifie l'etat des cases
			tmp.SetGrosCarre(list_pos[ligne][0]);
			tmp.SetGrosCarre(list_pos[ligne][1]);
			tmp.SetGrosCarre(list_pos[ligne][2]);
			tmp.SetGrosCarre(list_pos[ligne][3]);

			if(list_pos[ligne][0] == 13)
				tmp.SetFarWin(0);

			// enregistre la position
			map_output.insert(MapPair_t(tmp.GetHash(), tmp));

			// remet la grille a sa position d'origine
			tmp = *this;
		}
	}
}

void CGrille::GenerateRectangleHorizontal(MapGrille_t & map_output) const
{
	// une grille temporaire
	CGrille tmp;
	tmp = *this;

// la liste de toutes les positions possibles du rectangle horizontal 1X2
static int list_pos[][2] =
{ {  0,  1 },   {  1,  2 },   {  2,  3 },   {  4,  5 },
  {  5,  6 },   {  6,  7 },   {  8,  9 },   {  9, 10 },
  { 10, 11 },   { 12, 13 },   { 13, 14 },   { 14, 15 },
  { 16, 17 },   { 17, 18 },   { 18, 19 },
};

	// cherche pour toutes les positions possible si on peut mettre le rectangle
	for(int ligne = 0; ligne != sizeof(list_pos) / sizeof(list_pos[0]); ligne++)
	{
		if( (IsFree(list_pos[ligne][0]) == true) &&
			(IsFree(list_pos[ligne][1]) == true) )
		{
			// oui, on peut le mettre
			// modifie l'etat des cases
			tmp.SetRectangleHorizontal(list_pos[ligne][0]);
			tmp.SetRectangleHorizontal(list_pos[ligne][1]);

			// enregistre la position
			map_output.insert(MapPair_t(tmp.GetHash(), tmp));

			// remet la grille a sa position d'origine
			tmp = *this;
		}
	}
}

void CGrille::GenerateRectangleVertical(MapGrille_t & map_output) const
{
	// une grille temporaire
	CGrille tmp;
	tmp = *this;

// la liste de toutes les positions possibles du rectangle vertical 1X2
static int list_pos[][2] =
{ {  0,  4 },   {  1,  5 },   {  2,  6 },   {  3,  7 },
  {  4,  8 },   {  5,  9 },   {  6, 10 },   {  7, 11 },
  {  8, 12 },   {  9, 13 },   { 10, 14 },   { 11, 15 },
  { 12, 16 },   { 13, 17 },   { 14, 18 },   { 15, 19 },
};

	// cherche pour toutes les positions possible si on peut mettre le rectangle
	for(int ligne = 0; ligne != sizeof(list_pos) / sizeof(list_pos[0]); ligne++)
	{
		if( (IsFree(list_pos[ligne][0]) == true) &&
			(IsFree(list_pos[ligne][1]) == true) )
		{
			// oui, on peut le mettre
			// modifie l'etat des cases
			tmp.SetRectangleVertical(list_pos[ligne][0]);
			tmp.SetRectangleVertical(list_pos[ligne][1]);

			// enregistre la position
			map_output.insert(MapPair_t(tmp.GetHash(), tmp));

			// remet la grille a sa position d'origine
			tmp = *this;
		}
	}
}

void CGrille::GenerateMove(CLogger & /*Log*/, CListMouvement & ListMove)
{
	// une grille temporaire
	CGrille tmp(*this);

	// une table qui permet de savoir si la case est deja testee
	// initialise tous les elements de la table a false
	bool tested[20];
	memset(tested, 0, sizeof(tested));

	// cherche tous les mouvements possibles
	for(int pos = 0; pos != sizeof(m_case); pos++)
	{
		// si la case est deja testee, on continue
		if(tested[pos] == true)
			continue;

		// test si un gros carre
		if(m_case[pos] == 4)
		{
			//Log << "Gros carr\n";
			// check move up
			if( (IsFreeUp(pos) == true) && (IsFreeUp(pos + 1) == true) )
			{
				// move up is allowed
				// move the big square
				//Log << "Up is allowed for a big square\n";
				//tmp.Dump(Log);

				tmp.SetGrosCarre(pos - 4);
				tmp.SetGrosCarre(pos - 3);
				tmp.SetVide(pos + 4);
				tmp.SetVide(pos + 5);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveUp, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move down
			if( (IsFreeDown(pos + 4) == true) && (IsFreeDown(pos + 5) == true) )
			{
				// move down is allowed
				// move the big square
				//Log << "Down is allowed for a big square\n";
				//tmp.Dump(Log);

				tmp.SetGrosCarre(pos + 8);
				tmp.SetGrosCarre(pos + 9);
				tmp.SetVide(pos);
				tmp.SetVide(pos + 1);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveDown, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move left
			if( (IsFreeLeft(pos) == true) && (IsFreeLeft(pos + 4) == true) )
			{
				// move left is allowed
				// move the big square
				//Log << "Left is allowed for a big square\n";
				//tmp.Dump(Log);

				tmp.SetGrosCarre(pos - 1);
				tmp.SetGrosCarre(pos + 3);
				tmp.SetVide(pos + 1);
				tmp.SetVide(pos + 5);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveLeft, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move right
			if( (IsFreeRight(pos + 1) == true) && (IsFreeRight(pos + 5) == true) )
			{
				// move right is allowed
				// move the big square
				//Log << "Right is allowed for a big square\n";
				//tmp.Dump(Log);

				tmp.SetGrosCarre(pos + 2);
				tmp.SetGrosCarre(pos + 6);
				tmp.SetVide(pos);
				tmp.SetVide(pos + 4);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveRight, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// met a jour les cases testees
			tested[pos] = true;
			tested[pos + 1] = true;
			tested[pos + 4] = true;
			tested[pos + 5] = true;
			continue;
		}

		// test si rectangle horizontal
		if(m_case[pos] == 3)
		{
			//Log << "Rectangle horizontal\n";
			// check move up
			if( (IsFreeUp(pos) == true) && (IsFreeUp(pos + 1) == true) )
			{
				// move up is allowed
				// move the horizontal rectangle
				//Log << "Up is allowed for a horizontal rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleHorizontal(pos - 4);
				tmp.SetRectangleHorizontal(pos - 3);
				tmp.SetVide(pos);
				tmp.SetVide(pos + 1);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveUp, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move down
			if( (IsFreeDown(pos) == true) && (IsFreeDown(pos + 1) == true) )
			{
				// move down is allowed
				// move the horizontal rectangle
				//Log << "Down is allowed for a horizontal rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleHorizontal(pos + 4);
				tmp.SetRectangleHorizontal(pos + 5);
				tmp.SetVide(pos);
				tmp.SetVide(pos + 1);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveDown, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move left
			if(IsFreeLeft(pos) == true)
			{
				// move left is allowed
				// move the horizontal rectangle
				//Log << "Left is allowed for a horizontal rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleHorizontal(pos - 1);
				tmp.SetVide(pos + 1);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveLeft, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move right
			if(IsFreeRight(pos + 1) == true)
			{
				// move right is allowed
				// move the horizontal rectangle
				//Log << "Right is allowed for a horizontal rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleHorizontal(pos + 2);
				tmp.SetVide(pos);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveRight, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// met a jour les cases testees
			tested[pos] = true;
			tested[pos + 1] = true;
			continue;
		}

		// test si rectangle vertical
		if(m_case[pos] == 2)
		{
			//Log << "Rectangle vertical\n";
			// check move up
			if(IsFreeUp(pos) == true)
			{
				// move up is allowed
				// move the vertical rectangle
				//Log << "Up is allowed for a vertical rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleVertical(pos - 4);
				tmp.SetVide(pos + 4);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveUp, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move down
			if(IsFreeDown(pos + 4) == true)
			{
				// move down is allowed
				// move the vertical rectangle
				//Log << "Down is allowed for a vertical rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleVertical(pos + 8);
				tmp.SetVide(pos);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveDown, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move left
			if( (IsFreeLeft(pos) == true) && (IsFreeLeft(pos + 4) == true) )
			{
				// move left is allowed
				// move the vertical rectangle
				//Log << "Left is allowed for a vertical rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleVertical(pos - 1);
				tmp.SetRectangleVertical(pos + 3);
				tmp.SetVide(pos);
				tmp.SetVide(pos + 4);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveLeft, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move right
			if( (IsFreeRight(pos) == true) && (IsFreeRight(pos + 4) == true) )
			{
				// move right is allowed
				// move the vertical rectangle
				//Log << "Right is allowed for a vertical rectangle\n";
				//tmp.Dump(Log);

				tmp.SetRectangleVertical(pos + 1);
				tmp.SetRectangleVertical(pos + 5);
				tmp.SetVide(pos);
				tmp.SetVide(pos + 4);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveRight, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// met a jour les cases testees
			tested[pos] = true;
			tested[pos + 4] = true;
			continue;
		}

		// test si petit carre
		if(m_case[pos] == 1)
		{
			//Log << "Petit carr\n";
			// check move up
			if(IsFreeUp(pos) == true)
			{
				// move up is allowed
				// move the small square
				//Log << "Up is allowed for a small square\n";
				//tmp.Dump(Log);

				tmp.SetPetitCarre(pos - 4);
				tmp.SetVide(pos);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveUp, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move down
			if(IsFreeDown(pos) == true)
			{
				// move down is allowed
				// move the small square
				//Log << "Down is allowed for a small square\n";
				//tmp.Dump(Log);

				tmp.SetPetitCarre(pos + 4);
				tmp.SetVide(pos);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveDown, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move left
			if(IsFreeLeft(pos) == true)
			{
				// move left is allowed
				// move the small square
				//Log << "Left is allowed for a small square\n";
				//tmp.Dump(Log);

				tmp.SetPetitCarre(pos - 1);
				tmp.SetVide(pos);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveLeft, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// check move right
			if(IsFreeRight(pos) == true)
			{
				// move right is allowed
				// move the small square
				//Log << "Right is allowed for a small square\n";
				//tmp.Dump(Log);

				tmp.SetPetitCarre(pos + 1);
				tmp.SetVide(pos);
				//tmp.Dump(Log);

				// ajoute le mouvement possible  la liste des mouvement
				ListMove.Add(GetHash(), pos, CMouvement::MoveRight, tmp.GetHash());

				// reset the grid
				tmp = *this;
			}

			// met a jour les cases testees
			// pas trs utile pour une forme de 1 case mais restons coherent
			tested[pos] = true;
			continue;
		}

		// test si case vide
		if(m_case[pos] == 0)
		{
			// met a jour les cases testees
			// pas trs utile pour une forme de 1 case mais restons coherent
			tested[pos] = true;
			continue;
		}

		// type de case inconnue
		throw std::exception("Type de case inconnu");
	}
}

bool CGrille::IsFreeUp(const int pos) const
{
	// check upper border
	if(pos <= 3) return false;
	return IsFree(pos - 4);
}

bool CGrille::IsFreeDown(const int pos) const
{
	// check lower border
	if(pos >= 16) return false;
	return IsFree(pos + 4);
}

bool CGrille::IsFreeLeft(const int pos) const
{
	// check left border
	if(pos == 0) return false;
	if(pos == 4) return false;
	if(pos == 8) return false;
	if(pos == 12) return false;
	if(pos == 16) return false;
	return IsFree(pos - 1);
}

bool CGrille::IsFreeRight(const int pos) const
{
	// check left border
	if(pos == 3) return false;
	if(pos == 7) return false;
	if(pos == 11) return false;
	if(pos == 15) return false;
	if(pos == 19) return false;
	return IsFree(pos + 1);
}

void CGrille::_compute_hash(void)
{
	m_hash = 0;
	for(int boucle = 0; boucle != sizeof(m_case); boucle++)
	{
		Hash_t mask = m_case[boucle];
		mask <<= boucle * 3;
		m_hash |= mask;
	}
}
