// Copyright ram-0000
// Je donne mon accord  Developpez.com pour une ventuelle intgration dans les pages Sources du site.

#ifndef CGRILLE_INCLUDED
#define CGRILLE_INCLUDED

#include <vector>
#include <map>
#include <fstream>
#include "Logger.h"
#include "Mouvement.h"
#include "ListMouvement.h"

class CGrille
{
public:
	typedef std::map<Hash_t, CGrille> MapGrille_t;
	typedef std::pair<Hash_t, CGrille> MapPair_t;
	typedef MapGrille_t::const_iterator MapGrille_const_it_t;
	typedef MapGrille_t::iterator MapGrille_it_t;

	CGrille(void);
	inline CGrille(const CGrille & Other) { *this = Other; }
	CGrille & operator = (const CGrille & Other);

	bool operator == (const CGrille & Other) const;
	bool operator < (const CGrille & Other) const;

	// cree une grille vide
	void Vide(void);

	void Dump(CLogger & Log) const;
	std::string Dump(void) const;

	void GenerateGrosCarre(MapGrille_t & list_output) const;
	void GenerateRectangleHorizontal(MapGrille_t & list_output) const;
	void GenerateRectangleVertical(MapGrille_t & list_output) const;
	void GeneratePetitCarre(MapGrille_t & list_output) const;

	void GenerateMove(CLogger & Log, CListMouvement & ListMove);

	inline bool IsFree(const int pos) const { return m_case[pos] == 0; }

	inline void SetGrosCarre(const int pos) { m_case[pos] = 4; _compute_hash(); }
	inline void SetRectangleHorizontal(const int pos) { m_case[pos] = 3; _compute_hash(); }
	inline void SetRectangleVertical(const int pos) { m_case[pos] = 2; _compute_hash(); }
	inline void SetPetitCarre(const int pos) { m_case[pos] = 1; _compute_hash(); }
	inline void SetVide(const int pos) { m_case[pos] = 0; _compute_hash(); }

	bool IsFreeUp(const int pos) const;
	bool IsFreeDown(const int pos) const;
	bool IsFreeLeft(const int pos) const;
	bool IsFreeRight(const int pos) const;

	inline Hash_t GetHash(void) const { return m_hash; }

	bool Save(std::ofstream & OutputStream) const;
	bool Load(std::ifstream & InputStream);

	inline int GetFarWin(void) const { return m_far_win; }
	inline void SetFarWin(const int Val) { m_far_win = Val; }

	inline void SetMove(const CMouvement & Mouvement) { m_move_win = Mouvement; }
	inline const CMouvement & GetMove(void) const { return m_move_win; }

private:
	void _compute_hash(void);

	// une grille est composee de 20 cases de 5 lignes de 4 colonnes
	// pour des raisons de performance, elle est representee en interne par une seule ligne de 20 cases
	// si la case vaut 0, la case est libre, si la case vaut 1, elle est occupee
	unsigned char m_case[20];
	int m_far_win;
	CMouvement m_move_win;
	Hash_t m_hash;
};

#endif // #ifndef CGRILLE_INCLUDED