// Copyright ram-0000
// Je donne mon accord  Developpez.com pour une ventuelle intgration dans les pages Sources du site.

#include "ListGrille.h"
#include "MyTimer.h"
#include <sys/stat.h>
#include <algorithm>

CListGrille::CListGrille(void)
{
	m_filename = "DataGrille.bin";
}

CListGrille::~CListGrille(void)
{
	m_map.clear();
}

bool CListGrille::Save(CLogger & Log) const
{
	// lance un timer pour calculer le temps
	CMyTimer timer;
	timer.Start();

	// ouvre le fichier
	Log << "\nEcriture du fichier " << m_filename << "...\n";
	std::ofstream out(m_filename, std::ios_base::binary);
	if(out.is_open() == false)
	{
		//Log << "Erreur en ouvrant le fichier" << m_filename << "...\n";
		throw std::exception("Erreur en ouvrant le fichier");
	}

	// ecriture de la map
	// ------------------

	// nombre d'elements
	size_t size = m_map.size();
	out.write((char *)&size, sizeof(size));

	for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
	{
		it->second.Save(out);
	}

	// fermeture du fichier
	out.close();

	// affiche le temps de calcul
	timer.StopAndPrint(Log, "Temps mis pour crire le fichier");
	return true;
}

bool CListGrille::Load(CLogger & Log)
{
	// lance un timer pour calculer le temps
	CMyTimer timer;
	timer.Start();

	// ouvre le fichier
	Log << "\nChargement du fichier " << m_filename << "...\n";
	std::ifstream in(m_filename, std::ios_base::binary);
	if(in.is_open() == false)
	{
		new std::exception("Erreur en ouvrant le fichier");
		return false;
	}

	// lecture de la map
	// -----------------
	m_map.clear();

	// nombre d'elements
	int nb_item;
	in.read((char *)&nb_item, sizeof(nb_item));

	for(int boucle = 0; boucle != nb_item; boucle++)
	{
		CGrille tmp;
		tmp.Load(in);
		m_map.insert(CGrille::MapPair_t(tmp.GetHash(), tmp));
	}

	// fermeture du fichier
	in.close();

	// affiche le temps de calcul
	timer.StopAndPrint(Log, "Temps mis pour charger le fichier");
	return true;
}

bool CListGrille::Generate(CLogger & Log)
{
struct _stat stat_struct;
	if(_stat(m_filename, &stat_struct) == -1)
	{
		if(GenerateGrille(Log) == false)
			return false;

		if(GenerateMove(Log) == false)
			return false;

		if(FindBestPath(Log) == false)
			return false;

		if(Save(Log) == false)
			return false;
	}
	else
	{
		if(Load(Log) == false)
			return false;
	}

	return true;
}

bool CListGrille::GenerateGrille(CLogger & Log)
{
	// lance un timer pour calculer le temps
	CMyTimer timer;
	timer.Start();

	Log << "\nGnration des grilles...\n";

	// build the list and write the file
	// creation de la premiere liste de grilles, c'est une liste de 1 grille vide
	m_map.clear();
	CGrille tmp;
	tmp.Vide();
	m_map.insert(CGrille::MapPair_t(tmp.GetHash(), tmp));

	// genere la liste des grilles avec un gros carre
	CGrille::MapGrille_t new_map_grille;
	new_map_grille.clear();
	for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
	{
		it->second.GenerateGrosCarre(new_map_grille);
	}
	m_map = new_map_grille;
	//Log.Level(1); Log << "Il y a " << m_map.size() << " grilles avec 1 gros carr\n";

	int total_nb_rectangle = 5;
	for(int nb_rectangle = 0; nb_rectangle != total_nb_rectangle; nb_rectangle++)
	{
		// genere la liste des grilles avec un rectangle horizontal
		new_map_grille.clear();
		for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
		{
			it->second.GenerateRectangleHorizontal(new_map_grille);
		}
		for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
		{
			it->second.GenerateRectangleVertical(new_map_grille);
		}
		m_map = new_map_grille;
		new_map_grille.clear();
		//Log.Level(1); Log << "Il y a " << m_map.size() << " grilles avec 1 gros carr et " << nb_rectangle + 1 << (nb_rectangle == 0 ? " rectangle\n" : " rectangles\n");
	}

	int total_nb_petit_carre = 4;
	for(int nb_petit_carre = 0; nb_petit_carre != total_nb_petit_carre; nb_petit_carre++)
	{
		// genere la liste des grilles avec un carre
		new_map_grille.clear();
		for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
		{
			it->second.GeneratePetitCarre(new_map_grille);
		}
		m_map = new_map_grille;
		new_map_grille.clear();
		//Log.Level(1); Log << "Il y a " << m_map.size() << " grilles avec 1 gros carr et 5 rectangles et " << nb_petit_carre + 1 << (nb_petit_carre == 0 ? " petit carr\n" : " petit carrs\n");
	}

	// affiche le temps de calcul
	timer.StopAndPrint(Log, "Temps mis pour gnrer toutes les grilles");
	return true;
}

bool CListGrille::GenerateMove(CLogger & Log)
{
	// lance un timer pour calculer le temps
	CMyTimer timer;
	timer.Start();

	Log << "\nGnration de la liste des mouvements possibles...\n";

	for(CGrille::MapGrille_it_t it = m_map.begin(); it != m_map.end(); ++it)
	{
		// calcul la liste de toutes les grilles atteignables  partir de cette grille
		it->second.GenerateMove(Log, m_list_mouvement);
	}

	// affiche le temps de calcul
	timer.StopAndPrint(Log, "Temps mis pour calculer les mouvements possibles");

	return true;
}

bool CListGrille::FindBestPath(CLogger & Log)
{
	// lance un timer pour calculer le temps
	CMyTimer timer;
	timer.Start();

	Log << "\nRecherche des meilleurs chemins...\n";

	// fait la liste de toutes les grilles gagnantes
	std::vector<Hash_t> list_gagnant;
	for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
	{
		if(it->second.GetFarWin() == 0)
			list_gagnant.push_back(it->second.GetHash());
	}

	FindBestPath(Log, list_gagnant, 0);

	// affiche le temps de calcul
	timer.StopAndPrint(Log, "Temps mis pour calculer les meilleurs chemins");

	return true;
}

bool CListGrille::FindBestPath(CLogger & Log, std::vector<Hash_t> & ListReference, const int Level)
{
	// check if finished
	if(ListReference.size() == 0)
		return true;

	// lance un timer pour calculer le temps
	CMyTimer timer;
	timer.Start();

	//Log << "\nGnration de la liste des grilles de niveau " << Level + 1 << " (" << ListReference.size() << " rfrents)...\n";

	// get the list
	std::vector<Hash_t> next_level_list;
	for(size_t boucle = 0; boucle != ListReference.size(); boucle++)
	{
		std::vector<CMouvement> new_list;
		m_list_mouvement.GetListTo(ListReference[boucle], new_list);

		for(size_t boucle2 = 0; boucle2 != new_list.size(); boucle2++)
		{
			CGrille::MapGrille_it_t it = m_map.find(new_list[boucle2].GetHashFrom());
			if(it->second.GetFarWin() == -1)
			{
				it->second.SetFarWin(Level + 1);
				it->second.SetMove(new_list[boucle2]);
				next_level_list.push_back(new_list[boucle2].GetHashFrom());
			}
		}
	}

	// update the far
	ListReference = next_level_list;
	next_level_list.clear();

	// affiche le temps de calcul
	//timer.StopAndPrint(Log, "Temps mis pour calculer la liste des grilles");

	// recursive call
	FindBestPath(Log, ListReference, Level + 1);
	return true;
}

void CListGrille::DisplayStatGrille(CLogger & Log, const Hash_t HashGrille) const
{
	CGrille::MapGrille_const_it_t it = m_map.find(HashGrille);
	if(it == m_map.end())
	{
		Log << "Grille non trouve\n";
		return;
	}

	if(it->second.GetFarWin() == -1)
	{
		Log << "\nLa grille propose n'a pas de solution\n\n";
		return;
	}

	Log << "\nLa grille est gagnante en " << it->second.GetFarWin() << " coups\n\n";

	// get the list of grille
	std::vector<std::string> table_str;
	while(it->second.GetFarWin() != 0)
	{
		table_str.push_back(it->second.Dump());
		it = m_map.find(it->second.GetMove().GetHashTo());
		if(it == m_map.end())
		{
			Log << "Grille non trouve\n";
			return;
		}
	}
	table_str.push_back(it->second.Dump());

	//Display line by line
	int nb_per_line = 15;
	for(size_t boucle = 0; boucle < table_str.size(); boucle += nb_per_line)
	{
		display(Log, boucle, nb_per_line, table_str);
	}
}

void CListGrille::display(CLogger & Log, const size_t pos, const size_t nb, const std::vector<std::string> & tbl_str) const
{
	// ligne 1
	for(size_t boucle = 0; boucle != nb; boucle++)
	{
		if(pos + boucle >= tbl_str.size())
			Log << "     ";
		else
			Log << tbl_str[boucle + pos].substr(0, 4).c_str() << " ";
	}
	Log << "\n";

	// ligne 2
	for(size_t boucle = 0; boucle != nb; boucle++)
	{
		if(pos + boucle >= tbl_str.size())
			Log << "     ";
		else
			Log << tbl_str[boucle + pos].substr(4, 4).c_str() << " ";
	}
	Log << "\n";

	// ligne 3
	for(size_t boucle = 0; boucle != nb; boucle++)
	{
		if(pos + boucle >= tbl_str.size())
			Log << "     ";
		else
			Log << tbl_str[boucle + pos].substr(8, 4).c_str() << " ";
	}
	Log << "\n";

	// ligne 4
	for(size_t boucle = 0; boucle != nb; boucle++)
	{
		if(pos + boucle >= tbl_str.size())
			Log << "     ";
		else
			Log << tbl_str[boucle + pos].substr(12, 4).c_str() << " ";
	}
	Log << "\n";

	// ligne 5
	for(size_t boucle = 0; boucle != nb; boucle++)
	{
		if(pos + boucle >= tbl_str.size())
			Log << "     ";
		else
			Log << tbl_str[boucle + pos].substr(16, 4).c_str() << " ";
	}
	Log << "\n";
	Log << "\n";
}

void CListGrille::Stat(CLogger & Log)
{
	// nombre de grille total
	Log << "Il y a un total de " << m_map.size() << " grilles\n";

	// comptage des grilles gagnantes
	int longest = 0;
	int nb_gagnant = 0;
	int nb_perdant = 0;
	for(CGrille::MapGrille_const_it_t it = m_map.begin(); it != m_map.end(); ++it)
	{
		if(it->second.GetFarWin() == 0)
			nb_gagnant++;
		if(it->second.GetFarWin() == -1)
			nb_perdant++;
		if(it->second.GetFarWin() != -1)
		{
			if(it->second.GetFarWin() > longest)
				longest = it->second.GetFarWin();
		}
	}
	Log << "Il y a un total de " << nb_gagnant << " grilles directement gagnantes\n";
	Log << "Il y a un total de " << nb_perdant << " grilles sans solution\n";
	Log << "Le plus long chemin optimal est de " << longest << " mouvement\n";
}
